#include <cstddef>
using std::size_t;

#include <cstdlib>
using std::exit;

#include <cstdlib>
using std::atoi;

#include <cassert>

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

unsigned int n_rule = 2;
const bool periodic_bc = false;

const size_t L = 100;
const size_t L1 = L-1;
const double d = 0.75/L;
bool alive[L][L];
bool update[L][L];
const int delay = 50; // milliseconds

void display()
{
	glClear(GL_COLOR_BUFFER_BIT);
	for (size_t i = 0; i < L; ++i)
		for (size_t j = 0; j < L; ++j)
			if (alive[i][j])
			{
				const double x = 2.0*(i+0.5)/L - 1.0;
				const double y = 2.0*(j+0.5)/L - 1.0;	
				glBegin(GL_POLYGON);
					glVertex2f(x-d,y-d);
					glVertex2f(x-d,y+d);
					glVertex2f(x+d,y+d);
					glVertex2f(x+d,y-d);
				glEnd();
				glFlush();
			}
	glutSwapBuffers();
}

void keyboard(unsigned char key, int x, int y)
{
	if (key == 'q')
		exit(0);
}

void create_block(size_t i, size_t j)
{
	alive[i][j] = true;
	alive[(i+1)%L][j] = true;
	alive[i][(j+1)%L] = true;
	alive[(i+1)%L][(j+1)%L] = true;
}

void create_toad(size_t i, size_t j)
{
	alive[i][j] = true;
	alive[(i+1)%L][j] = true;
	alive[(i+2)%L][j] = true;
	alive[(i+1)%L][(j+1)%L] = true;
	alive[(i+2)%L][(j+1)%L] = true;
	alive[(i+3)%L][(j+1)%L] = true;
}

void create_boat(size_t i, size_t j)
{
	alive[(i+1)%L][j] = true;
	alive[i][(j+1)%L] = true;
	alive[(i+2)%L][(j+1)%L] = true;
	alive[i][(j+2)%L] = true;
	alive[(i+1)%L][(j+2)%L] = true;
}

void create_spaceship(size_t i, size_t j)
{
	alive[i][j] = true;
	alive[(i+1)%L][j] = true;
	alive[(i+2)%L][j] = true;
	alive[(i+3)%L][j] = true;
	alive[i][(j+1)%L] = true;
	alive[(i+4)%L][(j+1)%L] = true;
	alive[i][(j+2)%L] = true;
	alive[(i+1)%L][(j+3)%L] = true;
	alive[(i+4)%L][(j+3)%L] = true;
}

void create_blinker(size_t i, size_t j)
{
	alive[(L+i-1)%L][j] = true;
	alive[i][j] = true;
	alive[(i+1)%L][j] = true;
}

void create_NW_glider(size_t i, size_t j)
{
	alive[(i+1)%L][j] = true;
	alive[i][(j+1)%L] = true;
	alive[(L+i-1)%L][(j+1)%L] = true;
	alive[(L+i-1)%L][j] = true;
	alive[(L+i-1)%L][(L+j-1)%L] = true;
}

void initialize_grid(void)
{
	for (size_t i = 0; i < L; ++i)
		for (size_t j = 0; j < L; ++j)
			alive[i][j] = false;
	create_blinker(70,70);
	create_blinker(25,25);
	create_NW_glider(40,30);
	create_spaceship(20,20);
	create_boat(15,80);
	create_block(90,14);
	create_toad(70,50);
	create_NW_glider(50,50);
	create_blinker(30,70);
}

void sweep_grid(void)
{
	if (periodic_bc)
	{
		for (size_t i = 0; i < L; ++i)
			for (size_t j = 0; j < L; ++j)
			{
				unsigned int count = 0;
				if (alive[(i+1)%L][j]) ++count;
				if (alive[(i+1)%L][(j+1)%L]) ++count;
				if (alive[i][(j+1)%L]) ++count;
				if (alive[(L+i-1)%L][(j+1)%L]) ++count;
				if (alive[(L+i-1)%L][j]) ++count;
				if (alive[(L+i-1)%L][(L+j-1)%L]) ++count;
				if (alive[i][(L+j-1)%L]) ++count;
				if (alive[(i+1)%L][(L+j-1)%L]) ++count;
				assert(count >=0);
				assert(count <= 8);
				if (count == n_rule)
					update[i][j] = alive[i][j];
				else if (count == n_rule+1)
					update[i][j] = true;
				else
					update[i][j] = false;
			}
	}
	else
	{
		for (size_t i = 0; i < L; ++i)
			for (size_t j = 0; j < L; ++j)
			{
				unsigned int count = 0;
				if (i != L1 and alive[i+1][j]) ++count;
				if (i != L1 and j != L1 and alive[i+1][j+1]) ++count;
				if (j != L1 and alive[i][j+1]) ++count;
				if (i != 0 and j != L1 and alive[i-1][j+1]) ++count;
				if (i != 0 and alive[i-1][j]) ++count;
				if (i != 0 and j != 0 and alive[i-1][j-1]) ++count;
				if (j != 0 and alive[i][j-1]) ++count;
				if (i != L1 and j != 0 and alive[i+1][j-1]) ++count;
				assert(count >=0);
				assert(count <= 8);
				if (count == n_rule)
					update[i][j] = alive[i][j];
				else if (count == n_rule+1)
					update[i][j] = true;
				else
					update[i][j] = false;
			}

	}
	
	for (size_t i = 0; i < L; ++i)
		for (size_t j = 0; j < L; ++j)
			alive[i][j] = update[i][j];
}

void animate(int)
{
	sweep_grid();
	glutPostRedisplay();
	glutTimerFunc(delay,animate,0);
}

int main(int argc, char **argv)
{
	glutInit(&argc,argv);
	if (argc == 2)
		n_rule = atoi(argv[1]);
	else
		assert(argc == 1);
	assert (n_rule < 8);	
	glutInitDisplayMode (GLUT_RGB | GLUT_DOUBLE);
	glutInitWindowSize (640, 640);
	glutInitWindowPosition (100, 100);
	glutCreateWindow("The Game of Life");
	glutDisplayFunc(display);
	glutKeyboardFunc(keyboard);
	glutTimerFunc(delay,animate,0);
	
	initialize_grid();
	glutMainLoop();
}

